# coding: utf-8
"""
Export all SampleDB information readable by the user to an archive.
"""

import datetime
import io
import json
import os
import tarfile
import time
import typing
import zipfile

import flask

from .. import logic
from ..models import Permissions, ObjectLogEntryType


def get_export_infos(
        user_id: typing.Optional[int],
        object_ids: typing.Optional[typing.List[int]] = None,
        include_actions: bool = True,
        include_instruments: bool = True,
        include_users: bool = True,
        include_locations: bool = True,
        include_rdf_files: bool = True
) -> typing.Tuple[typing.Dict[str, typing.Union[bytes, str]], typing.Dict[str, typing.Any]]:
    archive_files: typing.Dict[str, typing.Union[bytes, str]] = {}
    if object_ids is None:
        relevant_instrument_ids = {
            instrument.id
            for instrument in logic.instruments.get_instruments()
        }
    else:
        relevant_instrument_ids = set()
    relevant_action_ids = set()
    relevant_user_ids = set()
    relevant_location_ids = set()
    relevant_markdown_images = set()
    objects = logic.object_permissions.get_objects_with_permissions(
        user_id=user_id,
        permissions=Permissions.READ,
        object_ids=object_ids
    )
    infos = {}
    object_infos: typing.List[typing.Dict[str, typing.Any]] = []
    for object in objects:
        if object_ids is not None and object.id not in object_ids:
            continue

        if include_rdf_files:
            archive_files[f"sampledb_export/{object.id}.rdf"] = logic.rdf.generate_rdf(user_id, object.id)

        object_infos.append({
            'id': object.id,
            'action_id': object.action_id,
            'versions': [],
            'comments': [],
            'location_assignments': [],
            'publications': [],
            'files': []
        })
        relevant_action_ids.add(object.action_id)

        for object_version in logic.objects.get_object_versions(object.id):
            relevant_user_ids.add(object_version.user_id)
            object_infos[-1]['versions'].append(
                {
                    'id': object_version.version_id,
                    'user_id': object_version.user_id,
                    'utc_datetime': object_version.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds') if object_version.utc_datetime is not None else None,
                    'schema': object_version.schema,
                    'data': object_version.data
                }
            )
            if object_version.schema:
                relevant_action_ids.update(logic.schemas.templates.find_used_template_ids(object_version.schema))

            for referenced_user_id, _ in logic.objects.find_user_references(object_version, False):
                relevant_user_ids.add(referenced_user_id)

            if object_version.data:
                for markdown in logic.markdown_to_html.get_markdown_from_object_data(object_version.data):
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(markdown)))

        for comment in logic.comments.get_comments_for_object(object.id):
            relevant_user_ids.add(comment.user_id)
            object_infos[-1]['comments'].append({
                'id': comment.id,
                'author_id': comment.user_id,
                'content': comment.content,
                'utc_datetime': comment.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds') if comment.utc_datetime is not None else None
            })

        for location_assignment in logic.locations.get_object_location_assignments(object.id):
            relevant_user_ids.add(location_assignment.user_id)
            if location_assignment.responsible_user_id:
                relevant_user_ids.add(location_assignment.responsible_user_id)
            if location_assignment.location_id:
                relevant_location_ids.add(location_assignment.location_id)
            if location_assignment.responsible_user_id:
                if location_assignment.confirmed:
                    status = 'confirmed'
                elif location_assignment.declined:
                    status = 'declined'
                else:
                    status = 'unconfirmed'
            else:
                status = None
            object_infos[-1]['location_assignments'].append({
                'id': location_assignment.id,
                'assigning_user_id': location_assignment.user_id,
                'responsible_user_id': location_assignment.responsible_user_id,
                'location_id': location_assignment.location_id,
                'utc_datetime': location_assignment.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds') if location_assignment.utc_datetime is not None else None,
                'status': status
            })

        log_entries = logic.object_log.get_object_log_entries(object.id, user_id)
        publication_log_entries = {
            (log_entry.data['doi'], log_entry.data['title']): log_entry
            for log_entry in log_entries
            if log_entry.type == ObjectLogEntryType.LINK_PUBLICATION
        }
        for publication_info in logic.publications.get_publications_for_object(object.id):
            object_infos[-1]['publications'].append({
                'doi': publication_info.doi,
                'title': publication_info.title,
                'object_name': publication_info.object_name
            })
            if (publication_info.doi, publication_info.title) in publication_log_entries:
                publication_log_entry = publication_log_entries[(publication_info.doi, publication_info.title)]
                object_infos[-1]['publications'][-1]['user_id'] = publication_log_entry.user_id
                object_infos[-1]['publications'][-1]['utc_datetime'] = publication_log_entry.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds')

        for file_info in logic.files.get_files_for_object(object.id):
            if not file_info.is_hidden:
                relevant_user_ids.add(file_info.user_id)
                object_infos[-1]['files'].append({
                    'id': file_info.id,
                    'hidden': False,
                    'title': file_info.title,
                    'description': file_info.description,
                    'uploader_id': file_info.user_id,
                    'utc_datetime': file_info.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds') if file_info.utc_datetime else None
                })
                if file_info.storage == 'database':
                    object_infos[-1]['files'][-1]['original_file_name'] = file_info.original_file_name
                    try:
                        file_bytes = file_info.open(read_only=True).read()
                    except Exception:
                        pass
                    else:
                        file_name = os.path.basename(file_info.original_file_name)
                        object_infos[-1]['files'][-1]['path'] = f'objects/{object.id}/files/{file_info.id}/{file_name}'
                        archive_files[f"sampledb_export/objects/{object.id}/files/{file_info.id}/{file_name}"] = file_bytes
                elif file_info.storage == 'url':
                    object_infos[-1]['files'][-1]['url'] = file_info.url
            else:
                object_infos[-1]['files'].append({
                    'id': file_info.id,
                    'hidden': True,
                    'hide_reason': file_info.hide_reason
                })
    infos['objects'] = object_infos

    if include_actions:
        action_infos = []
        for action_info in logic.action_permissions.get_actions_with_permissions(user_id=user_id, permissions=Permissions.READ):
            if action_info.id in relevant_action_ids:
                if action_info.user_id is not None:
                    relevant_user_ids.add(action_info.user_id)
                if action_info.instrument_id is not None:
                    relevant_instrument_ids.add(action_info.instrument_id)
                action_infos.append({
                    'id': action_info.id,
                    'type': action_info.type.object_name.get('en', 'object').lower() if action_info.type else 'object',
                    'name': action_info.name.get('en'),
                    'user_id': action_info.user_id,
                    'instrument_id': action_info.instrument_id if not flask.current_app.config['DISABLE_INSTRUMENTS'] else None,
                    'description': action_info.description.get('en'),
                    'description_is_markdown': action_info.description_is_markdown,
                    'short_description': action_info.short_description.get('en'),
                    'short_description_is_markdown': action_info.short_description_is_markdown
                })
                if action_info.description_is_markdown:
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(action_info.description.get('en', ''))))
                if action_info.short_description_is_markdown:
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(action_info.short_description.get('en', ''))))
        infos['actions'] = action_infos

    if include_instruments and not flask.current_app.config['DISABLE_INSTRUMENTS']:
        instrument_infos: typing.List[typing.Dict[str, typing.Any]] = []
        for instrument_info in logic.instruments.get_instruments():
            if instrument_info.id in relevant_instrument_ids:
                relevant_user_ids.update({user.id for user in instrument_info.responsible_users})
                instrument_infos.append({
                    'id': instrument_info.id,
                    'name': instrument_info.name.get('en'),
                    'description': instrument_info.description.get('en'),
                    'description_is_markdown': instrument_info.description_is_markdown,
                    'short_description': instrument_info.short_description.get('en'),
                    'short_description_is_markdown': instrument_info.short_description_is_markdown,
                    'instrument_scientist_ids': [user.id for user in instrument_info.responsible_users],
                    'instrument_log_entries': []
                })
                user_is_instrument_responsible = user_id in [user.id for user in instrument_info.responsible_users]
                if user_is_instrument_responsible:
                    instrument_infos[-1]['notes'] = instrument_info.notes.get('en')
                    instrument_infos[-1]['notes_is_markdown'] = instrument_info.notes_is_markdown
                if instrument_info.users_can_view_log_entries or user_is_instrument_responsible:
                    for log_entry in logic.instrument_log_entries.get_instrument_log_entries(instrument_info.id):
                        relevant_user_ids.add(log_entry.user_id)
                        instrument_infos[-1]['instrument_log_entries'].append({
                            'id': log_entry.id,
                            'author_id': log_entry.user_id,
                            'file_attachments': [],
                            'object_attachments': [],
                            'versions': []
                        })
                        for version in log_entry.versions:
                            instrument_infos[-1]['instrument_log_entries'][-1]['versions'].append({
                                'log_entry_id': log_entry.id,
                                'version_id': version.version_id,
                                'content': version.content,
                                'utc_datetime': version.utc_datetime.replace(tzinfo=None).isoformat(timespec='microseconds'),
                                'categories': []
                            })
                            for category in version.categories:
                                instrument_infos[-1]['instrument_log_entries'][-1]['versions'][-1]['categories'].append({
                                    'id': category.id,
                                    'title': category.title
                                })
                        for file_attachment in log_entry.file_attachments:
                            file_name = os.path.basename(file_attachment.file_name)
                            file_path = f'instruments/{instrument_info.id}/log_entries/{log_entry.id}/files/{file_attachment.id}/{file_name}'
                            instrument_infos[-1]['instrument_log_entries'][-1]['file_attachments'].append({
                                'file_name': file_attachment.file_name,
                                'path': file_path
                            })
                            archive_files["sampledb_export/" + file_path] = file_attachment.content

                        for object_attachment in log_entry.object_attachments:
                            instrument_infos[-1]['instrument_log_entries'][-1]['object_attachments'].append({
                                'object_id': object_attachment.object_id
                            })
                if instrument_info.description_is_markdown:
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(instrument_info.description.get('en', ''))))
                if instrument_info.short_description_is_markdown:
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(instrument_info.short_description.get('en', ''))))
                if user_is_instrument_responsible and instrument_info.notes_is_markdown:
                    relevant_markdown_images.update(logic.markdown_images.find_referenced_markdown_images(logic.markdown_to_html.markdown_to_safe_html(instrument_info.notes.get('en', ''))))
        infos['instruments'] = instrument_infos
    elif include_instruments:
        infos['instruments'] = []
    if include_users:
        user_infos = []
        for user_info in logic.users.get_users(exclude_hidden=False):
            if user_info.id in relevant_user_ids:
                if user_id is None:
                    user_infos.append({
                        'id': user_info.id,
                        'name': user_info.name
                    })
                else:
                    user_infos.append({
                        'id': user_info.id,
                        'name': user_info.name,
                        'orcid_id': f'https://orcid.org/{user_info.orcid}' if user_info.orcid else None,
                        'affiliation': user_info.affiliation if user_info.affiliation else None,
                        'role': user_info.role if user_info.role else None
                    })
        infos['users'] = user_infos

    locations = logic.locations.get_locations()
    all_relevant_location_ids: typing.Set[int] = set()
    while relevant_location_ids:
        new_relevant_location_ids: typing.Set[int] = set()
        for location_info in locations:
            if location_info.id in relevant_location_ids:
                if location_info.parent_location_id is not None:
                    new_relevant_location_ids.add(location_info.parent_location_id)
        all_relevant_location_ids.update(relevant_location_ids)
        relevant_location_ids = new_relevant_location_ids - all_relevant_location_ids

    readable_location_ids = [
        location.id
        for location in logic.location_permissions.get_locations_with_user_permissions(
            user_id=user_id,
            permissions=Permissions.READ
        )
    ]

    if include_locations:
        location_infos = []
        for location_info in locations:
            if location_info.id in all_relevant_location_ids:
                if location_info.id in readable_location_ids:
                    location_infos.append({
                        'id': location_info.id,
                        'parent_id': location_info.parent_location_id,
                        'name': location_info.name,
                        'description': location_info.description
                    })
                else:
                    location_infos.append({
                        'id': location_info.id,
                        'parent_id': location_info.parent_location_id,
                    })
        infos['locations'] = location_infos

    for file_name in relevant_markdown_images:
        content = logic.markdown_images.get_markdown_image(file_name, user_id)
        if content is not None:
            archive_files[f'sampledb_export/markdown_files/{file_name}'] = content
    return archive_files, infos


def get_archive_files(user_id: typing.Optional[int], object_ids: typing.Optional[typing.List[int]] = None) -> typing.Dict[str, bytes]:
    archive_files, infos = get_export_infos(user_id, object_ids)
    archive_files['sampledb_export/data.json'] = json.dumps(infos, indent=2)

    readme_text = f"""SampleDB Export
===============

The file data.json contains information on objects from the SampleDB instance at {flask.url_for('frontend.index', _external=True)} in JSON format. Aside from information on the objects themselves, it also contains information on relevant actions, instruments, users and locations.
"""
    if len(archive_files) > 1:
        readme_text += """
The objects directory contains files uploaded for the objects in data.json.
"""
    if user_id is not None:
        readme_text += f"""
This archive was created for user #{user_id} at {datetime.datetime.now(datetime.timezone.utc).isoformat(timespec='microseconds')}.
"""
    else:
        readme_text += f"""
This archive was created for an anonymous user at {datetime.datetime.now(datetime.timezone.utc).isoformat(timespec='microseconds')}.
"""

    archive_files["sampledb_export/README.txt"] = readme_text

    binary_archive_files = {}
    for file_name, file_content in archive_files.items():
        if isinstance(file_content, str):
            binary_archive_files[file_name] = file_content.encode('utf-8')
        else:
            binary_archive_files[file_name] = file_content

    return binary_archive_files


def get_zip_archive(
        user_id: typing.Optional[int],
        object_ids: typing.Optional[typing.List[int]] = None
) -> bytes:
    archive_files = get_archive_files(user_id, object_ids=object_ids)
    zip_bytes = io.BytesIO()
    with zipfile.ZipFile(zip_bytes, 'w') as zip_file:
        for file_name, file_content in archive_files.items():
            zip_file.writestr(file_name, file_content)
    return zip_bytes.getvalue()


def get_tar_gz_archive(
        user_id: typing.Optional[int],
        object_ids: typing.Optional[typing.List[int]] = None
) -> bytes:
    archive_files = get_archive_files(user_id, object_ids=object_ids)
    tar_bytes = io.BytesIO()
    with tarfile.open('sampledb_export.tar.gz', 'w:gz', fileobj=tar_bytes) as tar_file:
        for file_name, file_content in archive_files.items():
            tar_info = tarfile.TarInfo(file_name)
            tar_info.size = len(file_content)
            tar_info.mode = 0o444
            tar_info.mtime = int(time.time())
            tar_file.addfile(tar_info, fileobj=io.BytesIO(file_content))
    return tar_bytes.getvalue()


def get_eln_archive(user_id: int, object_ids: typing.Optional[typing.List[int]] = None) -> bytes:
    archive_files, infos = get_export_infos(
        user_id=user_id,
        object_ids=object_ids,
        include_rdf_files=False,
        include_users=True,
        include_actions=False,
        include_instruments=False,
        include_locations=False,
    )
    binary_archive_files = logic.eln_export.generate_ro_crate_metadata(archive_files, infos, user_id, object_ids)
    zip_bytes = io.BytesIO()
    with zipfile.ZipFile(zip_bytes, 'w') as zip_file:
        for file_name, file_content in binary_archive_files.items():
            zip_file.writestr(file_name, file_content)
    return zip_bytes.getvalue()


FILE_FORMATS = {
    '.zip': ('.zip Archive', get_zip_archive, 'application/zip'),
    '.tar.gz': ('.tar.gz Archive', get_tar_gz_archive, 'application/gzip'),
    '.eln': ('.eln File', get_eln_archive, 'application/zip')
}



# coding: utf-8
"""
Create default action type translations.
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..actions import ActionType
from ..languages import Language


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('action_types', 'name'):
        return False

    default_action_type_translation = [
        {
            'action_type_id': ActionType.SAMPLE_CREATION,
            'language_id': Language.ENGLISH,
            'name': 'Sample Creation',
            'description': 'These Actions represent processes which create a sample.',
            'object_name': 'Sample',
            'object_name_plural': 'Samples',
            'view_text': 'View Samples',
            'perform_text': 'Create Sample',
        },
        {
            'action_type_id': ActionType.MEASUREMENT,
            'language_id': Language.ENGLISH,
            'name': 'Measurement',
            'description': 'These Actions represent processes which perform a measurement.',
            'object_name': 'Measurement',
            'object_name_plural': 'Measurements',
            'view_text': 'View Measurements',
            'perform_text': 'Perform Measurement',
        },
        {
            'action_type_id': ActionType.SIMULATION,
            'language_id': Language.ENGLISH,
            'name': 'Simulation',
            'description': 'These Actions represent processes which run a simulation.',
            'object_name': 'Simulation',
            'object_name_plural': 'Simulations',
            'view_text': 'View Simulations',
            'perform_text': 'Run Simulation',
        },
        {
            'action_type_id': ActionType.SAMPLE_CREATION,
            'language_id': Language.GERMAN,
            'name': 'Probenerstellung',
            'description': 'Diese Aktionen repräsentieren Prozesse, die Proben erstellen.',
            'object_name': 'Probe',
            'object_name_plural': 'Proben',
            'view_text': 'Proben anzeigen',
            'perform_text': 'Probe erstellen',
        },
        {
            'action_type_id': ActionType.MEASUREMENT,
            'language_id': Language.GERMAN,
            'name': 'Messung',
            'description': 'Diese Aktionen repräsentieren Prozesse, die Messungen durchführen.',
            'object_name': 'Messung',
            'object_name_plural': 'Messungen',
            'view_text': 'Messungen anzeigen',
            'perform_text': 'Messung durchführen',
        },
        {
            'action_type_id': ActionType.SIMULATION,
            'language_id': Language.GERMAN,
            'name': 'Simulation',
            'description': 'Diese Aktionen repräsentieren Prozesse, die Simulationen durchführen.',
            'object_name': 'Simulation',
            'object_name_plural': 'Simulationen',
            'view_text': 'Simulationen anzeigen',
            'perform_text': 'Simulation durchführen',
        }
    ]
    existing_translation_action_type_and_language_ids = [
        (action_type_id, language_id)
        for action_type_id, language_id in db.session.execute(db.text("""
            SELECT action_type_id, language_id
            FROM action_type_translations
        """)).fetchall()
    ]
    performed_migration = False
    for action_type_translation in default_action_type_translation:
        # Skip migration by condition
        if (action_type_translation['action_type_id'], action_type_translation['language_id']) in existing_translation_action_type_and_language_ids:
            continue

        # Perform migration
        db.session.execute(db.text("""
            INSERT INTO action_type_translations (action_type_id, language_id, name, description, object_name, object_name_plural, view_text, perform_text)
            VALUES (:action_type_id, :language_id,:name, :description, :object_name, :object_name_plural, :view_text, :perform_text)
        """), params=action_type_translation)
        performed_migration = True
    return performed_migration



# coding: utf-8
"""
Replace the public_objects table with the all_user_object_permissions table.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    table_exists = db.session.execute(db.text("""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_name = 'public_objects'
    """)).fetchall()
    if not table_exists:
        return False

    # Perform migration
    public_objects = db.session.execute(db.text("""
        SELECT object_id
        FROM public_objects
    """)).fetchall()
    all_user_object_permissions_result = db.session.execute(db.text("""
        SELECT object_id, permissions
        FROM all_user_object_permissions
    """)).fetchall()
    all_user_object_permissions = {
        object_id: permissions
        for object_id, permissions in all_user_object_permissions_result
    }
    for object_id, in public_objects:
        if object_id not in all_user_object_permissions:
            db.session.execute(db.text("""
                INSERT INTO all_user_object_permissions
                (object_id, permissions)
                VALUES (:object_id, 'READ')
            """), {'object_id': object_id})
    view_exists = db.session.execute(db.text("""
        SELECT table_name
        FROM information_schema.views
        WHERE table_name = 'user_object_permissions_by_all'
    """)).fetchall()
    if view_exists:
        # the view will be recreated by the following migration, until then it
        # needs to be dropped as it likely depends on public_objects
        db.session.execute(db.text("""
            DROP VIEW user_object_permissions_by_all
        """))
    db.session.execute(db.text("""
        DROP TABLE public_objects
    """))
    return True



import datetime
import typing

import flask

from . import core
from ... import logic
from ...models import BackgroundTask, BackgroundTaskStatus, DataverseExport, DataverseExportStatus
from ... import db


def post_dataverse_export_task(
    object_id: int,
    user_id: int,
    server_url: str,
    api_token: str,
    dataverse: str,
    property_whitelist: typing.Optional[typing.Sequence[typing.List[typing.Union[str, int]]]],
    file_id_whitelist: typing.Sequence[int] = (),
    tag_whitelist: typing.Sequence[str] = ()
) -> typing.Tuple[typing.Union[BackgroundTaskStatus, typing.Tuple[bool, dict[str, typing.Any]]], typing.Optional[BackgroundTask]]:
    data = {
        'object_id': object_id,
        'user_id': user_id,
        'server_url': server_url,
        'api_token': api_token,
        'dataverse': dataverse,
        'property_whitelist': property_whitelist,
        'file_id_whitelist': file_id_whitelist,
        'tag_whitelist': tag_whitelist
    }
    if flask.current_app.config["ENABLE_BACKGROUND_TASKS"]:
        dataverse_export = DataverseExport(object_id, None, user_id, DataverseExportStatus.TASK_CREATED)
        db.session.add(dataverse_export)
        db.session.commit()
        return core.post_background_task(
            type='dataverse_export',
            data=data,
            auto_delete=False
        )
    else:
        return (handle_dataverse_export_task(data, None), None)  # type: ignore


def handle_dataverse_export_task(
    data: typing.Dict[str, typing.Any],
    task_id: typing.Optional[int]
) -> typing.Tuple[bool, typing.Optional[dict[str, typing.Any]]]:
    object_id: int = data['object_id']
    user_id: int = data['user_id']
    server_url: str = data['server_url']
    api_token: str = data['api_token']
    dataverse: str = data['dataverse']
    property_whitelist: typing.Optional[typing.Sequence[typing.List[typing.Union[str, int]]]] = data['property_whitelist']
    file_id_whitelist: typing.Sequence[int] = data['file_id_whitelist']
    tag_whitelist: typing.Sequence[str] = data['tag_whitelist']

    del data['api_token']
    if task_id:
        BackgroundTask.query.filter_by(id=task_id).update({"data": data})
        db.session.commit()

    credentials_are_valid = True
    validation_error = ""

    if not logic.dataverse_export.is_api_token_valid(server_url, api_token):
        credentials_are_valid = False
        validation_error = "API Token is not valid."
    else:
        if not logic.dataverse_export.dataverse_has_required_metadata_blocks(server_url, api_token, dataverse):
            credentials_are_valid = False
            validation_error = "The Dataverse does not have the required metadata blocks."

    if not credentials_are_valid:
        if task_id:
            expiration_date = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=30)

            db.session.delete(
                DataverseExport.query.filter_by(object_id=object_id).first()
            )
            db.session.commit()

            return False, {
                "error_message": validation_error,
                "expiration_date": expiration_date
            }
        else:
            return False, {
                "status": BackgroundTaskStatus.FAILED,
                "error_message": validation_error
            }
    else:
        try:
            success, url_or_error = logic.dataverse_export.upload_object(
                object_id=object_id,
                user_id=user_id,
                server_url=server_url,
                api_token=api_token,
                dataverse=dataverse,
                property_whitelist=property_whitelist,
                file_id_whitelist=file_id_whitelist,
                tag_whitelist=tag_whitelist
            )
        except logic.errors.DataverseNotReachableError:
            success = False
            url_or_error = '%(service_name)s cannot reach the %(dataverse_name)s API at this time. Please try again later.'

        result_key = "url" if success else "error_message"
        result: typing.Dict[str, typing.Any] = {
            result_key: url_or_error
        }
        if task_id:
            # background tasks enabled
            if not success:
                db.session.delete(
                    DataverseExport.query.filter_by(object_id=object_id).first()
                )
                db.session.commit()
            expiration_date = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=30)
            result['expiration_date'] = expiration_date

        return success, result



# coding: utf-8
"""
Add the disable_create_objects column to the action_types table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('action_types', 'is_template'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE action_types
        ADD is_template boolean NOT NULL DEFAULT(FALSE)
    """))
    return True



# coding: utf-8
"""

"""

import os
import flask

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

frontend = flask.Blueprint(
    'frontend',
    __name__,
    template_folder=os.path.join(os.path.abspath(os.path.dirname(__file__)), 'templates')
)


from . import actions
from . import action_types
from . import admin_warnings
from . import authentication_forms
from . import background_tasks
from . import dataverse_export
from . import download_service
from . import eln_import
from . import export
from . import favicon
from . import group_categories
from . import index
from . import info_pages
from . import instruments
from . import locations
from . import location_types
from . import markdown_images
from . import languages
from . import objects
from . import projects
from . import publications
from . import scicat_export
from . import status
from . import tags
from . import timezone
from . import topics
from . import users
from . import users_forms
from . import errors
from . import utils
from . import federation
from . import well_known


__all__ = [
    'actions',
    'action_types',
    'admin_warnings',
    'authentication_forms',
    'background_tasks',
    'dataverse_export',
    'download_service',
    'eln_import',
    'export',
    'favicon',
    'group_categories',
    'index',
    'info_pages',
    'instruments',
    'locations',
    'location_types',
    'markdown_images',
    'languages',
    'objects',
    'projects',
    'publications',
    'scicat_export',
    'status',
    'tags',
    'timezone',
    'topics',
    'users',
    'users_forms',
    'errors',
    'utils',
    'federation',
    'well_known',
]



import typing

import flask
from werkzeug.exceptions import HTTPException

from .authentication import http_basic_auth, http_api_token_or_access_token_auth, http_api_token_or_refresh_token_auth
from ...utils import FlaskResponseT


def is_api_error() -> bool:
    return flask.request.blueprint == 'api' or flask.request.blueprint is None and flask.request.path.startswith('/api/')


def handle_api_error(error: typing.Any) -> FlaskResponseT:
    return flask.jsonify({
        "message": getattr(error, 'description', "The server encountered an internal error and was unable to complete your request. Either the server is overloaded or there is an error in the application.")
    }), int(getattr(error, 'code', 500))


@http_basic_auth.error_handler
@http_api_token_or_access_token_auth.error_handler
@http_api_token_or_refresh_token_auth.error_handler
def auth_error(status: int) -> FlaskResponseT:
    try:
        flask.abort(status)
    except HTTPException as e:
        return handle_api_error(e)
    else:
        return handle_api_error(None)



from datetime import datetime
import typing

import flask

from .components import _get_or_create_component_id
from .utils import _get_id, _get_uuid, _get_utc_datetime, _get_str, _get_dict
from .users import _parse_user_ref, _get_or_create_user_id, UserRef
from ..comments import get_comment, create_comment, update_comment, Comment
from ..components import Component
from .. import errors, fed_logs, object_log
from ...models import Object


class CommentData(typing.TypedDict):
    fed_id: int
    component_uuid: str
    user: typing.Optional[UserRef]
    content: str
    utc_datetime: datetime


def parse_comment(
        comment_data: typing.Dict[str, typing.Any]
) -> CommentData:
    uuid = _get_uuid(comment_data.get('component_uuid'))
    fed_id = _get_id(comment_data.get('comment_id'))
    if uuid == flask.current_app.config['FEDERATION_UUID']:
        # do not accept updates for own data
        raise errors.InvalidDataExportError(f'Invalid update for local comment #{fed_id}')
    return CommentData(
        fed_id=fed_id,
        component_uuid=uuid,
        user=_parse_user_ref(_get_dict(comment_data.get('user'))),
        content=_get_str(comment_data.get('content'), mandatory=True, allow_empty=False),
        utc_datetime=_get_utc_datetime(comment_data.get('utc_datetime'), mandatory=True)
    )


def import_comment(
        comment_data: CommentData,
        object: Object,
        component: Component
) -> Comment:
    component_id = _get_or_create_component_id(comment_data['component_uuid'])
    assert component_id is not None
    # component_id will only be None if this would import a local comment

    user_id = _get_or_create_user_id(comment_data['user'])
    try:
        comment = get_comment(comment_data['fed_id'], component_id)

        if comment.user_id != user_id or comment.content != comment_data['content'] or comment.utc_datetime != comment_data['utc_datetime']:
            update_comment(
                comment_id=comment.id,
                user_id=user_id,
                content=comment_data['content'],
                utc_datetime=comment_data['utc_datetime']
            )
            fed_logs.update_comment(comment.id, component.id)
    except errors.CommentDoesNotExistError:
        assert component_id is not None
        comment = get_comment(create_comment(
            object_id=object.object_id,
            user_id=user_id,
            content=comment_data['content'],
            utc_datetime=comment_data['utc_datetime'],
            fed_id=comment_data['fed_id'],
            component_id=component_id
        ))
        fed_logs.import_comment(comment.id, component.id)
        if user_id is not None:
            object_log.post_comment(user_id=user_id, object_id=comment.object_id, comment_id=comment.id, utc_datetime=comment_data['utc_datetime'], is_imported=True)
    return comment


def parse_import_comment(
        comment_data: typing.Dict[str, typing.Any],
        object: Object,
        component: Component
) -> Comment:
    return import_comment(parse_comment(comment_data), object, component)



from functools import wraps
import json
import traceback
import typing

import cherrypy
import flask
from flask.views import MethodView
import werkzeug
import werkzeug.exceptions


ResponseContent = typing.Optional[typing.Union[typing.Dict[str, typing.Any], typing.Dict[int, typing.Any], typing.List[typing.Any], str, bool, int]]
ResponseData = typing.Union[werkzeug.Response, ResponseContent, typing.Tuple[ResponseContent], typing.Tuple[ResponseContent, int], typing.Tuple[ResponseContent, int, typing.Dict[str, str]]]


class Resource(MethodView):
    @classmethod
    def as_view(cls, name: typing.Any, *class_args: typing.Any, **class_kwargs: typing.Any) -> typing.Any:
        return super(Resource, _resource_decorator(cls)).as_view(name, *class_args, **class_kwargs)


def _resource_decorator(cls: typing.Type[Resource]) -> typing.Type[Resource]:
    for method in ['get', 'head', 'post', 'put', 'delete']:
        if hasattr(cls, method) and callable(getattr(cls, method)):
            setattr(cls, method, _resource_method_decorator(getattr(cls, method)))
    return cls


def _resource_method_decorator(f: typing.Callable[[typing.Any], ResponseData]) -> typing.Callable[[typing.Any], typing.Any]:
    @wraps(f)
    def decorator(*args: typing.Any, **kwargs: typing.Any) -> werkzeug.Response:
        flask.request.on_json_loading_failed = _on_json_loading_failed_replacement  # type: ignore
        try:
            response_data = f(*args, **kwargs)
            if isinstance(response_data, werkzeug.Response):
                response = response_data
            else:
                message: ResponseContent
                status = 200
                headers: typing.Dict[str, str] = {}
                if isinstance(response_data, tuple):
                    if len(response_data) == 1:
                        message, = response_data
                    elif len(response_data) == 2:
                        message, status = response_data
                    elif len(response_data) == 3:
                        message, status, headers = response_data
                    else:
                        message = None
                else:
                    message = response_data
                if message is None:
                    response = flask.current_app.response_class(
                        message,
                        status=status,
                        headers=headers
                    )
                else:
                    response = _make_json_response(
                        message,
                        status=status,
                        headers=headers
                    )
        except werkzeug.exceptions.HTTPException:
            raise
        except Exception as e:
            cherrypy.log(''.join(traceback.format_exception(e)))
            response = _make_json_response(
                obj={'message': 'Internal Server Error'},
                status=500
            )
        return response
    return decorator


def _make_json_response(
        obj: typing.Any,
        status: int = 200,
        headers: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> werkzeug.Response:
    if headers is None:
        headers = {}

    indent = None
    separators = (",", ":")

    compact_notation = getattr(flask.current_app.json, 'compact', None)
    if compact_notation is False or (compact_notation is None and flask.current_app.debug):
        indent = 2
        separators = (", ", ": ")

    return typing.cast(werkzeug.Response, flask.current_app.response_class(
        response=f"{json.dumps(obj=obj, indent=indent, separators=separators)}\n",
        mimetype=getattr(flask.current_app.json, 'mimetype', 'application/json'),
        status=status,
        headers=headers
    ))


def _on_json_loading_failed_replacement(_e: Exception) -> typing.NoReturn:
    flask.abort(_make_json_response(
        obj={'message': 'Failed to decode JSON object'},
        status=400
    ))



# coding: utf-8
"""
Add REFERENCED_BY_OBJECT_METADATA enum value to NotificationType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    return enum_value_migration('notificationtype', 'REFERENCED_BY_OBJECT_METADATA')



# coding: utf-8
"""
Script for building the SampleDB translations.

These are built automatically, unless the BUILD_TRANSLATIONS configuration
value is unset/set to False. This script mostly exists so that the
translations can be built once during testing instead of building them for
each individual test.

Usage: sampledb build_translations
"""
import sys
import typing

import sampledb


def main(arguments: typing.List[str]) -> None:
    if arguments:
        print(__doc__)
        sys.exit(1)
    sampledb.build_translations(sampledb.config.PYBABEL_PATH)
    print("Success: the translations have been built")



# coding: utf-8
"""
Replace NOT NULL constraints per column by NOT NULL constraints conditioned by federation reference in objects_previous.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    constraints = db.session.execute(db.text("""
             SELECT conname
             FROM pg_catalog.pg_constraint
             WHERE conname = 'objects_previous_not_null_check'
        """)).fetchall()
    if len(constraints) > 0:
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE objects_previous
            ADD CONSTRAINT objects_previous_not_null_check
                CHECK (
                    (
                        fed_object_id IS NOT NULL AND
                        fed_version_id IS NOT NULL AND
                        component_id IS NOT NULL
                    ) OR (
                        action_id IS NOT NULL AND
                        data IS NOT NULL AND
                        schema IS NOT NULL AND
                        user_id IS NOT NULL AND
                        utc_datetime IS NOT NULL
                    )
                ),
            ALTER COLUMN action_id DROP NOT NULL,
            ALTER COLUMN data DROP NOT NULL,
            ALTER COLUMN schema DROP NOT NULL,
            ALTER COLUMN user_id DROP NOT NULL,
            ALTER COLUMN utc_datetime DROP NOT NULL
    """))
    return True



# coding: utf-8
"""
Create default location types.
"""
import json

import flask_sqlalchemy

from .utils import table_has_column
from ..locations import LocationType


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    has_enable_instruments = table_has_column('location_types', 'enable_instruments')
    has_enable_capacities = table_has_column('location_types', 'enable_capacities')

    existing_location_type_ids = [
        location_type[0]
        for location_type in db.session.execute(db.text("""
                    SELECT id
                    FROM location_types;
                """)).fetchall()
    ]

    # checks if translatable columns still exist in action_types
    default_location_types = [
        {
            'id': LocationType.LOCATION,
            'name': json.dumps({'en': 'Locations', 'de': 'Orte'}),
            'location_name_singular': json.dumps({'en': 'Location', 'de': 'Ort'}),
            'location_name_plural': json.dumps({'en': 'Locations', 'de': 'Orte'}),
            'admin_only': False,
            'enable_parent_location': True,
            'enable_sub_locations': True,
            'enable_object_assignments': True,
            'enable_responsible_users': False,
            'show_location_log': False,
            'enable_instruments': True,
            'enable_capacities': False
        }
    ]

    performed_migration = False
    for location_type in default_location_types:
        # Skip migration by condition
        if location_type['id'] in existing_location_type_ids:
            continue

        # Perform migration
        if has_enable_capacities:
            # enable_instruments was added before enable_capacities
            assert has_enable_instruments
            db.session.execute(db.text("""
                INSERT INTO location_types (id, name, location_name_singular, location_name_plural, admin_only, enable_parent_location, enable_sub_locations, enable_object_assignments, enable_responsible_users, show_location_log, enable_instruments, enable_capacities)
                VALUES (:id, :name, :location_name_singular, :location_name_plural, :admin_only, :enable_parent_location, :enable_sub_locations, :enable_object_assignments, :enable_responsible_users, :show_location_log, :enable_instruments, :enable_capacities)
            """), params=location_type)
        elif has_enable_instruments:
            db.session.execute(db.text("""
                INSERT INTO location_types (id, name, location_name_singular, location_name_plural, admin_only, enable_parent_location, enable_sub_locations, enable_object_assignments, enable_responsible_users, show_location_log, enable_instruments)
                VALUES (:id, :name, :location_name_singular, :location_name_plural, :admin_only, :enable_parent_location, :enable_sub_locations, :enable_object_assignments, :enable_responsible_users, :show_location_log, :enable_instruments)
            """), params=location_type)
        else:
            db.session.execute(db.text("""
                INSERT INTO location_types (id, name, location_name_singular, location_name_plural, admin_only, enable_parent_location, enable_sub_locations, enable_object_assignments, enable_responsible_users, show_location_log)
                VALUES (:id, :name, :location_name_singular, :location_name_plural, :admin_only, :enable_parent_location, :enable_sub_locations, :enable_object_assignments, :enable_responsible_users, :show_location_log)
            """), params=location_type)
        performed_migration = True

    return performed_migration



# coding: utf-8
"""

"""

import flask
import flask_login

from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, SubmitField, SelectField, BooleanField
from wtforms.validators import Length, DataRequired, ValidationError, InputRequired, AnyOf


class CreateAPITokenForm(FlaskForm):
    description = StringField('description', validators=[Length(min=1, max=100)])


class AddOwnAPITokenForm(FlaskForm):
    token = StringField('token', validators=[Length(min=1, max=100)])
    description = StringField('description', validators=[Length(min=1, max=100)])


class AuthenticationMethodForm(FlaskForm):
    id = IntegerField('Authentication_method_id', validators=[DataRequired()])
    submit = SubmitField('Remove')


class AddComponentForm(FlaskForm):
    address = StringField(validators=[Length(min=0, max=100)])
    uuid = StringField(validators=[Length(min=1, max=100)])
    name = StringField(validators=[Length(min=0, max=100)])
    description = StringField()


class EditComponentForm(FlaskForm):
    address = StringField(validators=[Length(min=0, max=100)])
    name = StringField(validators=[Length(min=0, max=100)])
    description = StringField()


class SyncComponentForm(FlaskForm):
    ignore_last_sync_time = BooleanField(default=False)


class EditAliasForm(FlaskForm):
    component = IntegerField('Component', validators=[InputRequired()])
    name = StringField('Full Name')
    use_real_name = BooleanField('Use real name')
    use_real_email = BooleanField('Use real email')
    use_real_orcid = BooleanField('Use real ORCID iD')
    affiliation = StringField('Affiliation')
    use_real_affiliation = BooleanField('Use real affiliation')
    role = StringField('Role')
    use_real_role = BooleanField('Use real role')
    submit = SubmitField('Change Alias')

    def validate_name(self, field: StringField) -> None:
        if flask.current_app.config['ENFORCE_SPLIT_NAMES'] and flask_login.current_user.type.name.lower() == "person":
            name = field.data
            if name and ', ' not in name[1:-1]:
                raise ValidationError(_('Please enter your name as: surname, given names.'))


class AddAliasForm(FlaskForm):
    component = SelectField('Database', validators=[InputRequired()])
    name = StringField('Full Name')
    use_real_name = BooleanField('Use real name')
    use_real_email = BooleanField('Use real email')
    use_real_orcid = BooleanField('Use real ORCID iD')
    affiliation = StringField('Affiliation')
    use_real_affiliation = BooleanField('Use real affiliation')
    role = StringField('Role')
    use_real_role = BooleanField('Use real role')
    submit = SubmitField('Add Alias')

    def validate_name(self, field: StringField) -> None:
        if flask.current_app.config['ENFORCE_SPLIT_NAMES'] and flask_login.current_user.type.name.lower() == "person":
            name = field.data
            if name and ', ' not in name[1:-1]:
                raise ValidationError(_('Please enter your name as: surname, given names.'))


class DeleteAliasForm(FlaskForm):
    component = IntegerField('Component', validators=[InputRequired()])


class ModifyELNIdentityForm(FlaskForm):
    type = StringField(validators=[AnyOf(["remove", "revoke", "enable"])])
    eln_user_id = IntegerField(validators=[InputRequired()])


class FederatedUserCreationForm(FlaskForm):
    username = StringField(validators=[InputRequired()])

    def validate_username(self, field: StringField) -> None:
        if flask.current_app.config['ENFORCE_SPLIT_NAMES']:
            username = field.data
            if ', ' not in username[1:-1]:
                raise ValidationError(_("Please enter your name as: surname, given names."))



# coding: utf-8
"""
Add name_cache column to objects_current table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('objects_current', 'name_cache'):
        return False

    # Perform migration
    db.session.execute(db.text("""
    ALTER TABLE objects_current
    ADD COLUMN name_cache JSON NULL
    """))
    db.session.execute(db.text("""
    UPDATE objects_current
    SET name_cache = data -> 'name' -> 'text'
    """))
    return True



# coding: utf-8
"""
Splits action types into action_types and action_type_translations
"""

import flask_sqlalchemy

from .utils import table_has_column
from ..languages import Language


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not table_has_column('action_types', 'name'):
        return False

    # Perform migration
    existing_data = [
        action_data
        for action_data in db.session.execute(db.text("""
            SELECT id, name, description, object_name, object_name_plural, view_text, perform_text
            FROM action_types
        """)).fetchall()
    ]

    db.session.execute(db.text("""
            ALTER TABLE action_types
            DROP COLUMN name,
            DROP COLUMN object_name,
            DROP COLUMN object_name_plural,
            DROP COLUMN view_text,
            DROP COLUMN perform_text
        """))

    performed_migration = False
    for action_type_id, name, description, object_name, object_name_plural, view_text, perform_text in existing_data:
        translation = {
            'action_type_id': action_type_id,
            'language_id': Language.ENGLISH,
            'name': name,
            'description': description,
            'object_name': object_name,
            'object_name_plural': object_name_plural,
            'view_text': view_text,
            'perform_text': perform_text
        }
        # Perform migration
        db.session.execute(db.text("""
                      INSERT INTO action_type_translations (action_type_id, language_id, name, description, object_name, object_name_plural, view_text, perform_text)
                      VALUES (:action_type_id, :language_id, :name, :description, :object_name, :object_name_plural, :view_text, :perform_text)
                  """), params=translation)
        performed_migration = True

    return performed_migration



"""
Add fed_login_available column to components table
"""
import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column("components", "fed_login_available"):
        return False

    db.session.execute(db.text("""
        ALTER TABLE components
        ADD fed_login_available BOOLEAN NOT NULL
        DEFAULT FALSE;
    """))
    return True



# coding: utf-8
"""

"""
import typing

import flask
import flask_login
from flask_babel import _, gettext
from flask_wtf import FlaskForm
from wtforms.fields import StringField, BooleanField
from wtforms.validators import InputRequired

from . import frontend
from .. import logic
from .utils import check_current_user_is_not_readonly
from ..utils import FlaskResponseT


@frontend.route('/languages/')
@flask_login.login_required
def languages() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    return flask.render_template(
        'languages/languages.html',
        languages=logic.languages.get_languages()
    )


@frontend.route('/languages/<int(signed=True):language_id>', methods=['GET', 'POST'])
@flask_login.login_required
def language(language_id: int) -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)

    try:
        language = logic.languages.get_language(language_id)
    except logic.errors.LanguageDoesNotExistError:
        return flask.abort(404)

    mode = flask.request.args.get('mode', None)
    if mode == 'edit':
        return show_language_form(language_id)

    all_languages = logic.languages.get_languages()

    return flask.render_template(
        'languages/language.html',
        language=language,
        all_languages=all_languages
    )


@frontend.route('/language/new', methods=['GET', 'POST'])
@flask_login.login_required
def new_language() -> FlaskResponseT:
    if not flask_login.current_user.is_admin:
        return flask.abort(403)
    return show_language_form(None)


class LanguageForm(FlaskForm):
    name_english = StringField(validators=[InputRequired()])
    lang_code = StringField(validators=[InputRequired()])
    datetime_format_datetime = StringField(validators=[InputRequired()])
    datetime_format_moment = StringField(validators=[InputRequired()])
    datetime_format_moment_output = StringField(validators=[InputRequired()])
    date_format_moment_output = StringField(validators=[InputRequired()])
    enabled_for_input = BooleanField()
    enabled_for_user_interface = BooleanField()


def show_language_form(
        language_id: typing.Optional[int]
) -> FlaskResponseT:
    check_current_user_is_not_readonly()
    language_form = LanguageForm()

    if language_id is not None:
        try:
            language = logic.languages.get_language(language_id)
        except logic.errors.LanguageDoesNotExistError:
            return flask.abort(404)
        if 'language_submit' not in flask.request.form:
            language_form.name_english.data = language.names.get('en', '')
            language_form.lang_code.data = language.lang_code
            language_form.datetime_format_datetime.data = language.datetime_format_datetime
            language_form.datetime_format_moment.data = language.datetime_format_moment
            language_form.datetime_format_moment_output.data = language.datetime_format_moment_output
            language_form.date_format_moment_output.data = language.date_format_moment_output
            language_form.enabled_for_input.data = language.enabled_for_input
            language_form.enabled_for_user_interface.data = language.enabled_for_user_interface
    else:
        language = None

    if language_form.validate_on_submit():
        try:
            existing_language_for_code = logic.languages.get_language_by_lang_code(
                lang_code=language_form.lang_code.data.strip().lower()
            )
        except logic.errors.LanguageDoesNotExistError:
            existing_language_for_code = None
        names = {}
        for language in logic.languages.get_languages():
            key = 'name_' + str(language.id)
            value = flask.request.form.get(key, '').strip()
            if value:
                names[language.lang_code] = value
        names['en'] = language_form.name_english.data.strip()
        if language_id is None and existing_language_for_code is None:
            names[language_form.lang_code.data.strip().lower()] = flask.request.form.get('name_new', '').strip()
            language = logic.languages.create_language(
                names=names,
                lang_code=language_form.lang_code.data.strip().lower(),
                datetime_format_datetime=language_form.datetime_format_datetime.data.strip(),
                datetime_format_moment=language_form.datetime_format_moment.data.strip(),
                datetime_format_moment_output=language_form.datetime_format_moment_output.data.strip(),
                date_format_moment_output=language_form.date_format_moment_output.data.strip(),
                enabled_for_input=bool(language_form.enabled_for_input.data),
                enabled_for_user_interface=bool(language_form.enabled_for_user_interface.data)
            )
            return flask.redirect(flask.url_for('.language', language_id=language.id))
        elif language_id is None or (existing_language_for_code is not None and language_id != existing_language_for_code.id):
            flask.flash(_('This language code is already in use.'), 'error')
        elif language_id == logic.languages.Language.ENGLISH and not bool(language_form.enabled_for_input.data):
            flask.flash(_('The English language cannot be disabled for input.'), 'error')
        elif language_id == logic.languages.Language.ENGLISH and not bool(language_form.enabled_for_user_interface.data):
            flask.flash(_('The English language cannot be disabled for the user interface.'), 'error')
        elif language is not None and language.lang_code != language_form.lang_code.data and language.lang_code in logic.locale.SUPPORTED_LOCALES:
            flask.flash(_('The language code must stay "%(lang_code)s".', lang_code=language.lang_code), 'error')
        else:
            try:
                logic.languages.update_language(
                    language_id=language_id,
                    names=names,
                    lang_code=language_form.lang_code.data.strip().lower(),
                    datetime_format_datetime=language_form.datetime_format_datetime.data.strip(),
                    datetime_format_moment=language_form.datetime_format_moment.data.strip(),
                    datetime_format_moment_output=language_form.datetime_format_moment_output.data.strip(),
                    date_format_moment_output=language_form.date_format_moment_output.data.strip(),
                    enabled_for_input=bool(language_form.enabled_for_input.data),
                    enabled_for_user_interface=bool(language_form.enabled_for_user_interface.data)
                )
                return flask.redirect(flask.url_for('.language', language_id=language_id))
            except Exception:
                flask.flash(_('Something went wrong while updating the language. Please try again.'), 'error')

    return flask.render_template(
        'languages/language_form.html',
        language_form=language_form,
        existing_language=language,
        all_languages=logic.languages.get_languages(),
        keep_language_code=language is not None and language.lang_code in logic.locale.SUPPORTED_LOCALES,
        submit_text=gettext('Create') if language_id is None else gettext('Save')
    )



# coding: utf-8
"""
Add datetime_format_moment_output column to languages table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('languages', 'datetime_format_moment_output'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE languages
        ADD datetime_format_moment_output TEXT NOT NULL DEFAULT 'lll'
    """))
    db.session.execute(db.text("""
        UPDATE languages
        SET datetime_format_moment_output = 'MMM D, YYYY, h:mm:ss A'
        WHERE id = -99
    """))
    db.session.execute(db.text("""
        UPDATE languages
        SET datetime_format_moment_output = 'DD.MM.YYYY HH:mm:ss'
        WHERE id = -98
    """))
    return True



# coding: utf-8
"""
Script for listing all files with "local_reference" storage in SampleDB.

Usage: sampledb list_local_file_references
"""
import sys
import typing

from .. import create_app
from ..models.files import File


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 0:
        print(__doc__)
        sys.exit(1)
    app = create_app()
    with app.app_context():
        files = File.query.all()
        for file in files:
            if file.data is None:
                continue
            if file.data.get('storage') == 'local_reference':
                print(f" - object #{file.object_id} / file #{file.id}: {file.data.get('filepath')} ({'' if file.data.get('valid') else 'invalid, '}uploaded by user #{file.user_id})")



# coding: utf-8
"""
Add declined column to object location assignments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('object_location_assignments', 'declined'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ADD declined BOOLEAN DEFAULT FALSE NOT NULL
    """))
    db.session.execute(db.text("""
        ALTER TABLE object_location_assignments
        ADD CONSTRAINT object_location_assignments_not_both_states_check
            CHECK (
                NOT (confirmed AND declined)
            )
    """))
    return True



import base64
import binascii
import typing

from ..markdown_images import get_markdown_image
from ..components import Component
from .. import errors
from ...models import MarkdownImage
from ... import db


def parse_markdown_image(
        markdown_image_data: typing.Tuple[str, bytes],
        component: Component
) -> typing.Tuple[str, bytes]:
    filename, data = markdown_image_data
    try:
        md_image_data = base64.b64decode(data)
    except binascii.Error:
        raise errors.InvalidDataExportError(f'Invalid markdown image \'{filename}\'')
    return filename, md_image_data


def import_markdown_image(
        markdown_image_data: typing.Tuple[str, bytes],
        component: Component
) -> None:
    filename, data = markdown_image_data
    if get_markdown_image(filename, None, component.id) is None:
        md_image = MarkdownImage(filename, data, None, permanent=True, component_id=component.id)
        db.session.add(md_image)
        db.session.commit()


def parse_import_markdown_image(
        markdown_image_data: typing.Tuple[str, bytes],
        component: Component
) -> None:
    return import_markdown_image(parse_markdown_image(markdown_image_data, component), component)



# coding: utf-8
"""
This module defines functions which can help filtering objects by using the
sampledb.datetypes types. Each function returns a SQLAlchemy object which can
be used for the filter_func in
VersionedJSONSerializableObjectTables.get_current_objects().

Filters that use equality cannot be expected to be exact due to floating point
precision. These filters use operators that include a range of EPSILON of the
left operand's magnitude in base units.
"""

from datetime import datetime, date
import operator
import json
import typing

import flask
import sqlalchemy as db
from flask_login import current_user

from . import datatypes
from . import languages
from .units import get_old_dimensionality
from .utils import get_translated_text, get_postgres_timezone_alias
from ..models import Objects
from ..models.files import File
from ..models.file_log import FileLogEntry

EPSILON = 1e-7

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def float_operator_equals(left: typing.Any, right: typing.Any) -> typing.Any:
    return db.and_(
        left * (1 - db.func.sign(left) * EPSILON) <= right,
        left * (1 + db.func.sign(left) * EPSILON) >= right
    )


def float_operator_less_than_equals(left: typing.Any, right: typing.Any) -> typing.Any:
    return left * (1 - db.func.sign(left) * EPSILON) <= right


def float_operator_greater_than_equals(left: typing.Any, right: typing.Any) -> typing.Any:
    return left * (1 + db.func.sign(left) * EPSILON) >= right


def quantity_binary_operator(db_obj: typing.Any, other: datatypes.Quantity, operator: typing.Callable[[typing.Any, typing.Any], typing.Any]) -> typing.Any:
    return db.and_(
        db_obj['_type'].astext == 'quantity',
        db.or_(
            db_obj['dimensionality'].astext == str(other.dimensionality),
            db_obj['dimensionality'].astext == get_old_dimensionality(other.units),
        ),
        operator(db_obj['magnitude_in_base_units'].astext.cast(db.Float), other.magnitude_in_base_units)
    )


def quantity_equals(db_obj: typing.Any, other: datatypes.Quantity) -> typing.Any:
    return quantity_binary_operator(db_obj, other, float_operator_equals)


def quantity_less_than(db_obj: typing.Any, other: datatypes.Quantity) -> typing.Any:
    return quantity_binary_operator(db_obj, other, operator.lt)


def quantity_less_than_equals(db_obj: typing.Any, other: datatypes.Quantity) -> typing.Any:
    return quantity_binary_operator(db_obj, other, float_operator_less_than_equals)


def quantity_greater_than(db_obj: typing.Any, other: datatypes.Quantity) -> typing.Any:
    return quantity_binary_operator(db_obj, other, operator.gt)


def quantity_greater_than_equals(db_obj: typing.Any, other: datatypes.Quantity) -> typing.Any:
    return quantity_binary_operator(db_obj, other, float_operator_greater_than_equals)


def quantity_between(db_obj: typing.Any, left: datatypes.Quantity, right: datatypes.Quantity, including: bool = True) -> typing.Any:
    if left.dimensionality != right.dimensionality:
        return False
    if including:
        return db.and_(
            db_obj['_type'].astext == 'quantity',
            db.or_(
                db_obj['dimensionality'].astext == str(left.dimensionality),
                db_obj['dimensionality'].astext == get_old_dimensionality(left.units),
            ),
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) * (1 + db.func.sign(db_obj['magnitude_in_base_units'].astext.cast(db.Float)) * EPSILON) >= left.magnitude_in_base_units,
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) * (1 - db.func.sign(db_obj['magnitude_in_base_units'].astext.cast(db.Float)) * EPSILON) <= right.magnitude_in_base_units
        )
    else:
        return db.and_(
            db_obj['_type'].astext == 'quantity',
            db.or_(
                db_obj['dimensionality'].astext == str(left.dimensionality),
                db_obj['dimensionality'].astext == get_old_dimensionality(left.units),
            ),
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) > left.magnitude_in_base_units,
            db_obj['magnitude_in_base_units'].astext.cast(db.Float) < right.magnitude_in_base_units
        )


def datetime_binary_operator(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime], operator: typing.Callable[[typing.Any, date], typing.Any]) -> typing.Any:
    if isinstance(other, datatypes.DateTime):
        other = other.utc_datetime
    other_date = other.date()

    if flask.g.get('user') is not None and flask.g.user.timezone is not None:
        current_timezone_name = flask.g.user.timezone
    elif hasattr(current_user, 'timezone') and current_user.timezone is not None:
        current_timezone_name = current_user.timezone
    else:
        current_timezone_name = 'UTC'
    current_timezone_name = get_postgres_timezone_alias(current_timezone_name, other_date)
    naive_datetime = db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD HH24:MI:SS')
    local_timestamp = db.func.timezone(current_timezone_name, naive_datetime)
    locale_date = db.func.date_trunc('day', local_timestamp)
    return db.and_(
        db_obj['_type'].astext == 'datetime',
        operator(locale_date, other_date)
    )


def datetime_equals(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime]) -> typing.Any:
    return datetime_binary_operator(db_obj, other, operator.eq)


def datetime_less_than(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime]) -> typing.Any:
    return datetime_binary_operator(db_obj, other, operator.lt)


def datetime_less_than_equals(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime]) -> typing.Any:
    return datetime_binary_operator(db_obj, other, operator.le)


def datetime_greater_than(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime]) -> typing.Any:
    return datetime_binary_operator(db_obj, other, operator.gt)


def datetime_greater_than_equals(db_obj: typing.Any, other: typing.Union[datatypes.DateTime, datetime]) -> typing.Any:
    return datetime_binary_operator(db_obj, other, operator.ge)


def datetime_between(db_obj: typing.Any, left: typing.Union[datatypes.DateTime, datetime], right: typing.Union[datatypes.DateTime, datetime], including: bool = True) -> typing.Any:
    if isinstance(left, datatypes.DateTime):
        left = left.utc_datetime
    if isinstance(right, datatypes.DateTime):
        right = right.utc_datetime
    left_date = left.date()
    right_date = right.date()
    if including:
        return db.and_(
            db_obj['_type'].astext == 'datetime',
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') >= left_date,
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') <= right_date,
        )
    else:
        return db.and_(
            db_obj['_type'].astext == 'datetime',
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') > left_date,
            db.func.to_timestamp(db_obj['utc_datetime'].astext, 'YYYY-MM-DD') < right_date,
        )


def boolean_equals(db_obj: typing.Any, value: typing.Union[datatypes.Boolean, bool]) -> typing.Any:
    if isinstance(value, datatypes.Boolean):
        value = value.value
    return db.and_(
        db_obj['_type'].astext == 'bool',
        db_obj['value'].astext.cast(db.Boolean) == value
    )


def boolean_true(db_obj: typing.Any) -> typing.Any:
    return boolean_equals(db_obj, True)


def boolean_false(db_obj: typing.Any) -> typing.Any:
    return boolean_equals(db_obj, False)


def text_equals(db_obj: typing.Any, text: typing.Union[datatypes.Text, str]) -> typing.Any:
    if isinstance(text, datatypes.Text):
        text_str = get_translated_text(text.text)
    else:
        text_str = text
    return db.or_(
        db.and_(
            db_obj['_type'].astext == 'text',
            db.or_(
                db.and_(
                    db.func.jsonb_typeof(db_obj['text']) == 'string',
                    db_obj['text'].astext == text_str,
                ),
                *[
                    db.and_(
                        db_obj['text'].has_key(lang_code),
                        db_obj['text'][lang_code].astext == text_str
                    )
                    for lang_code in languages.get_language_codes()
                ]
            )
        ),
        db.and_(
            db_obj['_type'].astext == 'plotly_chart',
            db_obj['plotly']['layout']['title']['text'].astext == text_str
        )
    )


def text_contains(db_obj: typing.Any, text: typing.Union[datatypes.Text, str]) -> typing.Any:
    if isinstance(text, datatypes.Text):
        text_str = get_translated_text(text.text)
    else:
        text_str = text
    return db.or_(
        db.and_(
            db_obj['_type'].astext == 'text',
            db.or_(
                db.and_(
                    db.func.jsonb_typeof(db_obj['text']) == 'string',
                    db_obj['text'].astext.like('%' + text_str + '%')
                ),
                *[
                    db.and_(
                        db_obj['text'].has_key(lang_code),
                        db_obj['text'][lang_code].astext.like('%' + text_str + '%')
                    )
                    for lang_code in languages.get_language_codes()
                ]
            )
        ),
        db.and_(
            db_obj['_type'].astext == 'plotly_chart',
            db_obj['plotly']['layout']['title']['text'].astext.like('%' + text_str + '%')
        )
    )


def sample_equals(db_obj: typing.Any, object_id: int) -> typing.Any:
    return db.and_(
        db_obj['_type'].astext == 'sample',
        db_obj['object_id'].astext.cast(db.Integer) == object_id
    )


def reference_equals(db_obj: typing.Any, reference_id: int) -> typing.Any:
    return db.or_(
        db.and_(
            db.or_(
                db_obj['_type'].astext == 'object_reference',
                db_obj['_type'].astext == 'sample',
                db_obj['_type'].astext == 'measurement'
            ),
            db_obj['object_id'].astext.cast(db.Integer) == reference_id
        ),
        db.and_(
            db_obj['_type'].astext == 'user',
            db_obj['user_id'].astext.cast(db.Integer) == reference_id
        ),
    )


def tags_contain(db_obj: typing.Any, tag: str) -> typing.Any:
    tag = tag.strip().lower()
    return db.and_(
        db_obj['_type'].astext == 'tags',
        db_obj['tags'].contains(json.dumps(tag))
    )


def attribute_not_set(db_obj: typing.Any) -> typing.Any:
    return db_obj == db.null()


def _has_file(db_obj: typing.Any, file_filter: db.ColumnElement[bool]) -> typing.Any:
    matching_files = db.select(
        File.object_id
    ).distinct().outerjoin(
        FileLogEntry,
        db.and_(
            FileLogEntry.object_id == File.object_id,
            FileLogEntry.file_id == File.id,
        )
    ).where(
        file_filter
    ).group_by(
        File.object_id,
        File.id
    ).having(
        db.func.sum(  # pylint: disable=not-callable
            db.case(
                {
                    'HIDE_FILE': 1,
                    'UNHIDE_FILE': -1
                },
                else_=0,
                value=FileLogEntry.type
            )
        ) == 0
    ).subquery()
    if hasattr(db_obj, 'object_id_column'):
        object_id_column = db_obj.object_id_column
    else:
        object_id_column = Objects.object_id_column
    return db.and_(
        object_id_column == matching_files.c.object_id,
    )


def file_name_contains(db_obj: typing.Any, text: typing.Union[datatypes.Text, str]) -> typing.Any:
    if isinstance(text, datatypes.Text):
        text_str = get_translated_text(text.text)
    else:
        text_str = text
    return _has_file(
        db_obj,
        db.or_(
            File.data.op("->>")("url").cast(db.Text).like('%' + text_str + '%'),
            File.data.op("->>")("original_file_name").cast(db.Text).like('%' + text_str + '%'),
            File.data.op("->>")("filepath").cast(db.Text).like('%' + text_str + '%')
        )
    )


def file_name_equals(db_obj: typing.Any, text: typing.Union[datatypes.Text, str]) -> typing.Any:
    if isinstance(text, datatypes.Text):
        text_str = get_translated_text(text.text)
    else:
        text_str = text
    return _has_file(
        db_obj,
        db.or_(
            File.data.op("->>")("url").cast(db.Text) == text_str,
            File.data.op("->>")("original_file_name").cast(db.Text) == text_str,
            File.data.op("->>")("filepath").cast(db.Text) == text_str
        )
    )



from . import action_types
from . import actions
from . import comments
from . import components
from . import files
from . import instruments
from . import location_types
from . import locations
from . import login
from . import markdown_images
from . import object_location_assignments
from . import objects
from . import update
from . import users
from . import utils

__all__ = [
    'action_types',
    'actions',
    'comments',
    'components',
    'files',
    'instruments',
    'location_types',
    'locations',
    'login',
    'markdown_images',
    'object_location_assignments',
    'objects',
    'update',
    'users',
    'utils',
]



import base64
import datetime
import typing

import flask

from .utils import _get_id, _get_uuid, _get_str, _get_dict, _get_list, _get_utc_datetime, _get_permissions
from .users import _parse_user_ref, _get_or_create_user_id, UserRef
from .actions import _parse_action_ref, _get_or_create_action_id, schema_entry_preprocessor, _parse_schema, ActionRef
from .comments import import_comment, parse_comment, CommentData
from .files import import_file, parse_file, FileData
from .object_location_assignments import import_object_location_assignment, parse_object_location_assignment, ObjectLocationAssignmentData
from .locations import LocationRef
from ..files import get_files_for_object
from ..markdown_images import find_referenced_markdown_images, get_markdown_image
from ..object_permissions import set_user_object_permissions, set_group_object_permissions, set_project_object_permissions, set_object_permissions_for_all_users, object_permissions
from ..schemas import validate_schema, validate
from ..schemas.validate import validate_eln_urls
from ..actions import get_action
from ..comments import get_comments_for_object
from ..components import get_component_by_uuid, get_component, Component
from ..groups import get_group
from ..locations import get_location, get_object_location_assignments
from ..objects import get_fed_object, get_object, update_object_version, insert_fed_object_version, get_object_versions
from ..object_log import create_object, edit_object
from ..projects import get_project
from ..users import get_user, check_user_exists
from .. import errors, fed_logs, languages, markdown_to_html
from ...models import Permissions, Object
from ...models.file_log import FileLogEntry, FileLogEntryType


class ObjectVersionData(typing.TypedDict):
    fed_version_id: int
    user: typing.Optional[UserRef]
    data: typing.Optional[typing.Dict[str, typing.Any]]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    utc_datetime: typing.Optional[datetime.datetime]
    import_notes: typing.List[str]


class ObjectPermissionsData(typing.TypedDict):
    users: typing.Dict[int, Permissions]
    groups: typing.Dict[int, Permissions]
    projects: typing.Dict[int, Permissions]
    all_users: Permissions


class ObjectData(typing.TypedDict):
    fed_object_id: int
    component_id: int
    action: typing.Optional[ActionRef]
    versions: typing.List[ObjectVersionData]
    comments: typing.List[CommentData]
    files: typing.List[FileData]
    object_location_assignments: typing.List[ObjectLocationAssignmentData]
    permissions: ObjectPermissionsData
    sharing_user: typing.Optional[UserRef]


class SharedFileHideData(typing.TypedDict):
    user: UserRef
    reason: str
    utc_datetime: datetime.datetime


class SharedFileData(typing.TypedDict):
    file_id: int
    component_uuid: str
    user: typing.Optional[UserRef]
    data: typing.Optional[typing.Dict[str, typing.Any]]
    utc_datetime: typing.Optional[str]
    hidden: typing.Optional[SharedFileHideData]


class SharedCommentData(typing.TypedDict):
    comment_id: int
    component_uuid: str
    user: typing.Optional[UserRef]
    content: typing.Optional[str]
    utc_datetime: datetime.datetime


class SharedObjectVersionData(typing.TypedDict):
    version_id: int
    user: typing.Optional[UserRef]
    data: typing.Optional[typing.Dict[str, typing.Any]]
    schema: typing.Optional[typing.Dict[str, typing.Any]]
    utc_datetime: typing.Optional[str]


class SharedObjectLocationAssignmentData(typing.TypedDict):
    id: int
    component_uuid: str
    location: typing.Optional[LocationRef]
    responsible_user: typing.Optional[UserRef]
    user: typing.Optional[UserRef]
    description: typing.Optional[typing.Dict[str, str]]
    utc_datetime: typing.Optional[str]
    confirmed: bool
    declined: bool


class SharedObjectData(typing.TypedDict):
    object_id: int
    component_uuid: str
    action: typing.Optional[ActionRef]
    versions: typing.List[SharedObjectVersionData]
    comments: typing.List[SharedCommentData]
    files: typing.List[SharedFileData]
    object_location_assignments: typing.List[SharedObjectLocationAssignmentData]
    policy: typing.Dict[str, typing.Any]
    sharing_user: typing.Optional[UserRef]


def import_object(
        object_data: ObjectData,
        component: Component,
        *,
        import_status: typing.Optional[typing.Dict[str, typing.Any]] = None
) -> Object:
    action_id = _get_or_create_action_id(object_data['action'])
    object = None
    all_import_notes = []
    did_perform_import = False
    for version in object_data['versions']:
        try:
            object = get_fed_object(object_data['fed_object_id'], component.id, version['fed_version_id'])
        except errors.ObjectDoesNotExistError:
            object = None
        except errors.ObjectVersionDoesNotExistError:
            object = None

        user_id = _get_or_create_user_id(version['user'])
        sharing_user_id = _get_or_create_user_id(object_data.get('sharing_user'))

        if object is None:
            object = insert_fed_object_version(
                fed_object_id=object_data['fed_object_id'],
                fed_version_id=version['fed_version_id'],
                component_id=component.id,
                data=version['data'],
                schema=version['schema'],
                user_id=user_id,
                action_id=action_id,
                utc_datetime=version['utc_datetime'],
                allow_disabled_languages=True,
                get_missing_schema_from_action=False  # if the version contains None for the schema, do not try to load it from the action
            )
            if object:
                fed_logs.import_object(object.id, component.id, version.get('import_notes', []), sharing_user_id, version['fed_version_id'])
                if user_id is not None:
                    if version['fed_version_id'] == 0:
                        create_object(
                            user_id=user_id,
                            object_id=object.id,
                            utc_datetime=version['utc_datetime'],
                            is_imported=True
                        )
                    else:
                        edit_object(
                            user_id=user_id,
                            object_id=object.id,
                            version_id=object.version_id,
                            utc_datetime=version['utc_datetime'],
                            is_imported=True
                        )
                did_perform_import = True
        elif object.schema != version['schema'] or object.data != version['data'] or object.user_id != user_id or object.action_id != action_id or object.utc_datetime != version['utc_datetime']:
            object = update_object_version(
                object_id=object.object_id,
                version_id=object.version_id,
                data=version['data'],
                schema=version['schema'],
                user_id=user_id,
                action_id=action_id,
                utc_datetime=version['utc_datetime'],
                allow_disabled_languages=True,
                get_missing_schema_from_action=False  # if the version contains None for the schema, do not try to load it from the action
            )
            fed_logs.update_object(object.id, component.id, version.get('import_notes', []), sharing_user_id, version['fed_version_id'])
            did_perform_import = True
        all_import_notes.extend(version.get('import_notes', []))
    if object is None:
        object = get_fed_object(
            fed_object_id=object_data['fed_object_id'],
            component_id=object_data['component_id']
        )

    for comment_data in object_data['comments']:
        import_comment(comment_data, object, component)

    for file_data in object_data['files']:
        import_file(file_data, object, component)

    for assignment_data in object_data['object_location_assignments']:
        import_object_location_assignment(assignment_data, object, component)

    # apply policy permissions as a minimum, but do not reduce existing permissions
    current_permissions_for_users = object_permissions.get_permissions_for_users(resource_id=object.object_id)
    for user_id, permission in object_data['permissions']['users'].items():
        if permission not in current_permissions_for_users.get(user_id, Permissions.NONE):
            set_user_object_permissions(object.object_id, user_id, permission)

    current_permissions_for_groups = object_permissions.get_permissions_for_groups(resource_id=object.object_id)
    for group_id, permission in object_data['permissions']['groups'].items():
        if permission not in current_permissions_for_groups.get(group_id, Permissions.NONE):
            set_group_object_permissions(object.object_id, group_id, permission)

    current_permissions_for_projects = object_permissions.get_permissions_for_projects(resource_id=object.object_id)
    for project_id, permission in object_data['permissions']['projects'].items():
        if permission not in current_permissions_for_projects.get(project_id, Permissions.NONE):
            set_project_object_permissions(object.object_id, project_id, permission)

    current_permissions_for_all_users = object_permissions.get_permissions_for_all_users(resource_id=object.object_id)
    permission = object_data['permissions']['all_users']
    if permission not in current_permissions_for_all_users:
        set_object_permissions_for_all_users(object.object_id, permission)

    if import_status is not None and did_perform_import:
        import_status['success'] = True
        import_status['notes'] = all_import_notes
        import_status['object_id'] = object.object_id
        import_status['utc_datetime'] = datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S')

    return object


def parse_object(
        object_data: typing.Dict[str, typing.Any],
        component: Component
) -> ObjectData:
    fed_object_id = _get_id(object_data.get('object_id'))
    versions = _get_list(object_data.get('versions'), mandatory=True)
    parsed_versions = []
    for version in versions:
        import_notes = []
        fed_version_id = _get_id(version.get('version_id'), mandatory=True, min=0)
        data: typing.Optional[typing.Dict[str, typing.Any]] = _get_dict(version.get('data'))
        schema: typing.Optional[typing.Dict[str, typing.Any]] = _get_dict(version.get('schema'))
        if schema is None or schema is None:
            data = None
            schema = None
        else:
            parse_entry(data, component)
            _parse_schema(schema)
            try:
                if schema is not None:
                    validate_schema(schema, strict=False)
                    if data is not None:
                        validate(data, schema, allow_disabled_languages=True, strict=False)
            except errors.ValidationError as e:
                schema = None
                data = None
                import_notes.append(f'Invalid data or schema in version {fed_version_id} of object #{fed_object_id} @ {component.uuid} ({e})')
        parsed_versions.append(ObjectVersionData(
            fed_version_id=fed_version_id,
            data=data,
            schema=schema,
            user=_parse_user_ref(_get_dict(version.get('user'))),
            utc_datetime=_get_utc_datetime(version.get('utc_datetime'), default=None),
            import_notes=import_notes
        ))

    result = ObjectData(
        fed_object_id=fed_object_id,
        component_id=component.id,
        versions=parsed_versions,
        action=_parse_action_ref(_get_dict(object_data.get('action'))),
        comments=[],
        files=[],
        object_location_assignments=[],
        permissions=ObjectPermissionsData(
            users={},
            groups={},
            projects={},
            all_users=Permissions.NONE
        ),
        sharing_user=_parse_user_ref(_get_dict(object_data.get('sharing_user')))
    )

    comments = _get_list(object_data.get('comments'))
    if comments is not None:
        for comment in comments:
            result['comments'].append(parse_comment(comment))

    files = _get_list(object_data.get('files'))
    if files is not None:
        for file in files:
            result['files'].append(parse_file(file))

    assignments = _get_list(object_data.get('object_location_assignments'))
    if assignments is not None:
        for assignment in assignments:
            result['object_location_assignments'].append(parse_object_location_assignment(assignment))

    policy = _get_dict(object_data.get('policy'), mandatory=True)

    permissions = _get_permissions(policy.get('permissions'))
    if permissions is not None:
        users = _get_dict(permissions.get('users'))
        if users is not None:
            for user_id, permission in users.items():
                user_id = _get_id(user_id, convert=True)
                try:
                    check_user_exists(user_id)
                except errors.UserDoesNotExistError:
                    continue
                try:
                    result['permissions']['users'][user_id] = Permissions.from_name(permission)
                except ValueError:
                    raise errors.InvalidDataExportError(f'Unknown permission "{permission}"')
        groups = _get_dict(permissions.get('groups'))
        if groups is not None:
            for group_id, permission in groups.items():
                group_id = _get_id(group_id, convert=True)
                try:
                    get_group(group_id)
                except errors.GroupDoesNotExistError:
                    continue
                try:
                    result['permissions']['groups'][group_id] = Permissions.from_name(permission)
                except ValueError:
                    raise errors.InvalidDataExportError(f'Unknown permission "{permission}"')
        projects = _get_dict(permissions.get('projects'))
        if projects is not None:
            for project_id, permission in projects.items():
                project_id = _get_id(project_id, convert=True)
                try:
                    get_project(project_id)
                except errors.ProjectDoesNotExistError:
                    continue
                try:
                    result['permissions']['projects'][project_id] = Permissions.from_name(permission)
                except ValueError:
                    raise errors.InvalidDataExportError(f'Unknown permission "{permission}"')
        all_users = _get_str(permissions.get('all_users'))
        if all_users is not None:
            permission = all_users
            try:
                result['permissions']['all_users'] = Permissions.from_name(permission)
            except ValueError:
                raise errors.InvalidDataExportError(f'Unknown permission "{permission}"')
    return result


def parse_import_object(
        object_data: typing.Dict[str, typing.Any],
        component: Component
) -> Object:
    return import_object(parse_object(object_data, component), component)


def shared_object_preprocessor(
        object_id: int,
        policy: typing.Dict[str, typing.Any],
        refs: typing.List[typing.Tuple[str, int]],
        markdown_images: typing.Dict[str, str],
        *,
        sharing_user_id: typing.Optional[int] = None
) -> SharedObjectData:
    result = SharedObjectData(
        object_id=object_id,
        versions=[],
        component_uuid=flask.current_app.config['FEDERATION_UUID'],
        action=None,
        comments=[],
        object_location_assignments=[],
        files=[],
        policy=policy,
        sharing_user=None
    )
    if sharing_user_id is not None:
        if ('users', sharing_user_id) not in refs:
            refs.append(('users', sharing_user_id))
        sharing_user = get_user(sharing_user_id)
        if sharing_user.fed_id is not None and sharing_user.component is not None:
            result['sharing_user'] = {
                'user_id': sharing_user.fed_id,
                'component_uuid': sharing_user.component.uuid
            }
        else:
            result['sharing_user'] = {
                'user_id': sharing_user.id,
                'component_uuid': flask.current_app.config['FEDERATION_UUID']
            }
    object = get_object(object_id)
    object_versions = get_object_versions(object_id)
    if 'access' in policy:
        if 'action' in policy['access'] and policy['access']['action'] and object.action_id is not None:
            if ('actions', object.action_id) not in refs:
                refs.append(('actions', object.action_id))
            action = get_action(object.action_id)
            if action.component is not None and action.fed_id is not None:
                comp = action.component
                result['action'] = ActionRef(
                    action_id=action.fed_id,
                    component_uuid=comp.uuid
                )
            else:
                result['action'] = ActionRef(
                    action_id=action.id,
                    component_uuid=flask.current_app.config['FEDERATION_UUID']
                )
        if 'comments' in policy['access'] and policy['access']['comments']:
            comments = get_comments_for_object(object_id)
            result['comments'] = []
            for comment in comments:
                if comment.component is None or comment.fed_id is None:
                    res_comment = {
                        'comment_id': comment.id,
                        'component_uuid': flask.current_app.config['FEDERATION_UUID']
                    }
                else:
                    comp = comment.component
                    res_comment = {
                        'comment_id': comment.fed_id,
                        'component_uuid': comp.uuid
                    }
                res_comment['content'] = comment.content
                res_comment['utc_datetime'] = comment.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if comment.utc_datetime is not None else None
                if 'users' in policy['access'] and policy['access']['users'] and comment.user_id is not None:
                    if ('users', comment.user_id) not in refs:
                        refs.append(('users', comment.user_id))
                    user = get_user(comment.user_id)
                    if user.fed_id is not None and user.component is not None:
                        comp = user.component
                        res_comment['user'] = {
                            'user_id': user.fed_id,
                            'component_uuid': comp.uuid
                        }
                    else:
                        res_comment['user'] = {
                            'user_id': comment.user_id,
                            'component_uuid': flask.current_app.config['FEDERATION_UUID']
                        }
                result['comments'].append(SharedCommentData(
                    comment_id=res_comment['comment_id'],
                    component_uuid=res_comment['component_uuid'],
                    user=res_comment.get('user'),
                    content=res_comment['content'],
                    utc_datetime=res_comment['utc_datetime']
                ))
        if 'files' in policy['access'] and policy['access']['files']:
            files = get_files_for_object(object_id)
            result['files'] = []
            for file in files:
                if file.component_id is None:
                    res_file = {
                        'file_id': file.id,
                        'component_uuid': flask.current_app.config['FEDERATION_UUID']
                    }
                else:
                    comp = get_component(file.component_id)
                    res_file = {
                        'file_id': file.fed_id,
                        'component_uuid': comp.uuid
                    }
                res_file['utc_datetime'] = file.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if file.utc_datetime else None
                if 'users' in policy['access'] and policy['access']['users'] and file.user_id is not None:
                    if ('users', file.user_id) not in refs:
                        refs.append(('users', file.user_id))
                    user = get_user(file.user_id)
                    if user.fed_id is not None and user.component is not None:
                        comp = user.component
                        res_file['user'] = {
                            'user_id': user.fed_id,
                            'component_uuid': comp.uuid
                        }
                    else:
                        res_file['user'] = {
                            'user_id': file.user_id,
                            'component_uuid': flask.current_app.config['FEDERATION_UUID']
                        }
                if file.is_hidden:
                    log_entry = FileLogEntry.query.filter_by(
                        object_id=file.object_id,
                        file_id=file.id,
                        type=FileLogEntryType.HIDE_FILE
                    ).order_by(FileLogEntry.utc_datetime.desc()).first()
                    res_file['hidden'] = {
                        'reason': file.hide_reason,
                        'utc_datetime': log_entry.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if log_entry is not None else None
                    }
                    if log_entry is not None:
                        if ('users', log_entry.user_id) not in refs:
                            refs.append(('users', log_entry.user_id))
                    if file.user_id:
                        user = get_user(file.user_id)
                        if user.fed_id is not None and user.component is not None:
                            comp = user.component
                            res_file['hidden']['user'] = {
                                'user_id': user.fed_id,
                                'component_uuid': comp.uuid
                            }
                        else:
                            res_file['hidden']['user'] = {
                                'user_id': file.user_id,
                                'component_uuid': flask.current_app.config['FEDERATION_UUID']
                            }
                    else:
                        res_file['hidden']['user'] = None
                else:
                    res_file['data'] = file.data
                    if file.storage == 'database':
                        res_file['data']['storage'] = 'federation'

                result['files'].append(SharedFileData(
                    file_id=res_file['file_id'],
                    component_uuid=res_file['component_uuid'],
                    user=res_file.get('user'),
                    data=res_file.get('data'),
                    utc_datetime=res_file['utc_datetime'],
                    hidden=res_file.get('hidden')
                ))
        if 'object_location_assignments' in policy['access'] and policy['access']['object_location_assignments']:
            olas = get_object_location_assignments(object_id)
            for ola in olas:
                if ola.location_id is not None and ('locations', ola.location_id) not in refs:
                    refs.append(('locations', ola.location_id))
                if ola.user_id is not None and ('users', ola.user_id) not in refs:
                    refs.append(('users', ola.user_id))
                if ola.responsible_user_id is not None and ('users', ola.responsible_user_id) not in refs:
                    refs.append(('users', ola.responsible_user_id))
                if ola.location_id is not None:
                    location = get_location(ola.location_id)
                    if location.component is not None and location.fed_id is not None:
                        comp = location.component
                        location_ref = LocationRef(
                            location_id=location.fed_id,
                            component_uuid=comp.uuid
                        )
                    else:
                        location_ref = LocationRef(
                            location_id=location.id,
                            component_uuid=flask.current_app.config['FEDERATION_UUID']
                        )
                else:
                    location_ref = None
                if ola.responsible_user_id is not None:
                    responsible_user = get_user(ola.responsible_user_id)
                    if responsible_user.component is not None and responsible_user.fed_id is not None:
                        comp = responsible_user.component
                        responsible_user_ref = UserRef(
                            user_id=responsible_user.fed_id,
                            component_uuid=comp.uuid
                        )
                    else:
                        responsible_user_ref = UserRef(
                            user_id=responsible_user.id,
                            component_uuid=flask.current_app.config['FEDERATION_UUID']
                        )
                else:
                    responsible_user_ref = None
                if ola.user_id is not None:
                    ola_user = get_user(ola.user_id)
                    if ola_user.component is not None and ola_user.fed_id is not None:
                        comp = ola_user.component
                        c_user = UserRef(
                            user_id=ola_user.fed_id,
                            component_uuid=comp.uuid
                        )
                    else:
                        c_user = UserRef(
                            user_id=ola_user.id,
                            component_uuid=flask.current_app.config['FEDERATION_UUID']
                        )
                else:
                    c_user = None
                if ola.component_id is None:
                    component_uuid = flask.current_app.config['FEDERATION_UUID']
                else:
                    comp = get_component(ola.component_id)
                    component_uuid = comp.uuid
                result['object_location_assignments'].append(SharedObjectLocationAssignmentData(
                    id=ola.id,
                    component_uuid=component_uuid,
                    location=location_ref,
                    responsible_user=responsible_user_ref,
                    user=c_user,
                    description=ola.description,
                    utc_datetime=ola.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if ola.utc_datetime else None,
                    confirmed=ola.confirmed,
                    declined=ola.declined,
                ))
    for version in object_versions:
        version_data: typing.Optional[typing.Dict[str, typing.Any]] = None
        version_schema: typing.Optional[typing.Dict[str, typing.Any]] = None
        version_user: typing.Optional[UserRef] = None
        if policy.get('access', {'data': True}).get('data', True):
            if version.data is not None:
                version_data = version.data.copy()
                entry_preprocessor(version_data, refs, markdown_images)
            if version.schema is not None:
                version_schema = version.schema.copy()
                schema_entry_preprocessor(version_schema, refs)
        if policy.get('access', {}).get('users', False) and version.user_id:
            if ('users', version.user_id) not in refs:
                refs.append(('users', version.user_id))
            user = get_user(version.user_id)
            if user.component is not None and user.fed_id is not None:
                comp = user.component
                version_user = UserRef(
                    user_id=user.fed_id,
                    component_uuid=comp.uuid
                )
            else:
                version_user = UserRef(
                    user_id=user.id,
                    component_uuid=flask.current_app.config['FEDERATION_UUID']
                )
        if 'modification' in policy:
            modification = policy['modification']
            if 'insert' in modification:
                if 'data' in modification['insert'] and version_data is not None:
                    for key in modification['insert']['data']:
                        version_data[key] = modification['insert']['data'][key]
                if 'schema' in modification['insert'] and version_schema is not None:
                    for key in modification['insert']['schema']:
                        version_schema['properties'][key] = modification['insert']['schema'][key]

            if 'update' in modification:
                if 'data' in modification['update'] and version_data is not None:
                    for key in modification['update']['data']:
                        if key not in version_data:
                            pass
                        for k in modification['update']['data'][key]:
                            version_data[key][k] = modification['update']['data'][key][k]
                if 'schema' in modification['update'] and version_schema is not None:
                    for key in modification['update']['schema']:
                        if key not in version_schema['properties']:
                            pass
                        version_schema['properties'][key] = {
                            key: value
                            for key, value in modification['update']['schema'][key].items()
                        }
        result['versions'].append(SharedObjectVersionData(
            version_id=version.version_id,
            schema=version_schema,
            data=version_data,
            user=version_user,
            utc_datetime=version.utc_datetime.strftime('%Y-%m-%d %H:%M:%S.%f') if version.utc_datetime else None
        ))
    return result


def entry_preprocessor(
        data: typing.Any,
        refs: typing.List[typing.Tuple[str, int]],
        markdown_images: typing.Dict[str, str]
) -> None:
    if type(data) is list:
        for entry in data:
            entry_preprocessor(entry, refs, markdown_images)
    elif type(data) is dict:
        if '_type' not in data.keys():
            for key in data:
                entry_preprocessor(data[key], refs, markdown_images)
        else:
            if data['_type'] in ['sample', 'object_reference', 'measurement']:
                if (data.get('component_uuid') is None or data.get('component_uuid') == flask.current_app.config['FEDERATION_UUID']) and 'eln_source_url' not in data:
                    o = get_object(data['object_id'])
                    if o.component_id is not None:
                        c = get_component(o.component_id)
                        data['object_id'] = o.fed_object_id
                        data['component_uuid'] = c.uuid
                    else:
                        data['component_uuid'] = flask.current_app.config['FEDERATION_UUID']
            if data['_type'] == 'user':
                if (data.get('component_uuid') is None or data.get('component_uuid') == flask.current_app.config['FEDERATION_UUID']) and 'eln_source_url' not in data:
                    try:
                        u = get_user(data.get('user_id'))  # type: ignore
                        if u.component_id is not None:
                            c = get_component(u.component_id)
                            data['user_id'] = u.fed_id
                            data['component_uuid'] = c.uuid
                        else:
                            data['component_uuid'] = flask.current_app.config['FEDERATION_UUID']
                    except errors.UserDoesNotExistError:
                        pass
            if data['_type'] == 'file':
                if data.get('component_uuid') is None:
                    data['component_uuid'] = flask.current_app.config['FEDERATION_UUID']

            if data['_type'] == 'text' and data.get('is_markdown'):
                if isinstance(data.get('text'), str):
                    markdown_as_html = markdown_to_html.markdown_to_safe_html(data['text'])
                    for file_name in find_referenced_markdown_images(markdown_as_html):
                        markdown_image_b = get_markdown_image(file_name, None)
                        data['text'] = data['text'].replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                        if markdown_image_b is not None and file_name not in markdown_images:
                            markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')
                elif isinstance(data.get('text'), dict):
                    for lang in data['text'].keys():
                        markdown_as_html = markdown_to_html.markdown_to_safe_html(data['text'][lang])
                        for file_name in find_referenced_markdown_images(markdown_as_html):
                            markdown_image_b = get_markdown_image(file_name, None)
                            data['text'][lang] = data['text'][lang].replace('/markdown_images/' + file_name, '/markdown_images/' + flask.current_app.config['FEDERATION_UUID'] + '/' + file_name)
                            if markdown_image_b is not None and file_name not in markdown_images:
                                markdown_images[file_name] = base64.b64encode(markdown_image_b).decode('utf-8')


def parse_entry(
        entry_data: typing.Any,
        component: Component
) -> None:
    if type(entry_data) is list:
        for entry in entry_data:
            parse_entry(entry, component)
    elif type(entry_data) is dict:
        if '_type' not in entry_data.keys():
            for key in entry_data:
                parse_entry(entry_data[key], component)
        else:
            if entry_data.get('_type') == 'user':
                user_id = _get_id(entry_data.get('user_id'), mandatory=False)
                if user_id is None:
                    return
                if 'component_uuid' not in entry_data and 'eln_source_url' in entry_data:
                    try:
                        validate_eln_urls(entry_data, [])
                    except errors.ValidationError:
                        raise errors.InvalidDataExportError('Invalid .eln file URLs')
                elif _get_uuid(entry_data.get('component_uuid')) == flask.current_app.config['FEDERATION_UUID']:
                    try:
                        user = get_user(user_id)
                    except errors.UserDoesNotExistError:
                        raise errors.InvalidDataExportError(f'Local user #{user_id} does not exist')
                    del entry_data['component_uuid']
                    entry_data['user_id'] = user.id
                else:
                    try:
                        data_component = get_component_by_uuid(_get_uuid(entry_data.get('component_uuid')))
                        try:
                            user = get_user(_get_id(entry_data.get('user_id')), data_component.id)
                            del entry_data['component_uuid']
                            entry_data['user_id'] = user.id
                        except errors.UserDoesNotExistError:
                            pass
                    except errors.ComponentDoesNotExistError:
                        pass

            if entry_data.get('_type') in ('sample', 'object_reference', 'measurement'):
                data_obj_id = _get_id(entry_data.get('object_id'), mandatory=False)
                if data_obj_id is None:
                    return
                if 'component_uuid' not in entry_data and 'eln_source_url' in entry_data:
                    try:
                        validate_eln_urls(entry_data, [])
                    except errors.ValidationError:
                        raise errors.InvalidDataExportError('Invalid .eln file URLs')
                elif _get_uuid(entry_data.get('component_uuid')) == flask.current_app.config['FEDERATION_UUID']:
                    try:
                        obj = get_object(data_obj_id)
                    except errors.ObjectDoesNotExistError:
                        raise errors.InvalidDataExportError(f'Local object #{data_obj_id} does not exist')
                    del entry_data['component_uuid']
                    entry_data['object_id'] = obj.id
                else:
                    try:
                        data_component = get_component_by_uuid(_get_uuid(entry_data.get('component_uuid')))
                        obj = get_fed_object(data_obj_id, data_component.id)
                        if obj is not None:
                            del entry_data['component_uuid']
                            entry_data['object_id'] = obj.object_id
                    except errors.ComponentDoesNotExistError:
                        pass
                    except errors.ObjectDoesNotExistError:
                        pass
            all_lang_codes = languages.get_language_codes()

            if entry_data.get('_type') == 'text':
                text_data = entry_data.get('text')
                if isinstance(text_data, dict):
                    for lang_code in list(text_data.keys()):
                        if lang_code not in all_lang_codes:
                            del entry_data['text'][lang_code]



# coding: utf-8
"""
Add IMPORT_FROM_ELN_FILE enum value to ObjectLogEntryType enum.
"""

import flask_sqlalchemy

from .utils import enum_value_migration


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    return enum_value_migration('objectlogentrytype', 'IMPORT_FROM_ELN_FILE')



# coding: utf-8
"""

"""

import typing
import copy
from .errors import ObjectDoesNotExistError
from .objects import get_object
from .object_permissions import get_user_object_permissions
from ..models import Permissions


def get_notebook_templates(
        object_id: int,
        data: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        user_id: int
) -> typing.List[typing.Dict[str, typing.Any]]:
    """
    Return notebook templates with parameters from object data.

    :param object_id: the ID of the object
    :param data: the data of the object
    :param schema: the schema of the object
    :param user_id: the user for which the notebook templates are prepared.
    :return: the prepared notebook template data
    """
    notebook_templates: typing.List[typing.Dict[str, typing.Any]] = copy.deepcopy(schema.get('notebookTemplates', []))
    for notebook_template in notebook_templates:
        for parameter, parameter_value in list(notebook_template.get('params', {}).items()):
            if parameter_value == 'object_id':
                notebook_template['params'][parameter] = object_id
            elif isinstance(parameter_value, list):
                parameter_data: typing.Optional[typing.Dict[str, typing.Any]] = data
                parameter_schema: typing.Optional[typing.Dict[str, typing.Any]] = schema
                for step in parameter_value:
                    if isinstance(step, str) and isinstance(parameter_data, dict) and step in parameter_data:
                        parameter_data = parameter_data[step]
                        if parameter_schema is not None and parameter_schema.get('type') == 'object' and step in parameter_schema.get('properties', {}):
                            parameter_schema = parameter_schema['properties'][step]
                        else:
                            parameter_schema = None
                    elif isinstance(step, int) and isinstance(parameter_data, list) and isinstance(parameter_schema, dict) and parameter_schema.get('type') == 'array' and 0 <= step < len(parameter_data):
                        parameter_data = parameter_data[step]
                        parameter_schema = parameter_schema.get('items', {})
                    elif step == 'object' and isinstance(parameter_data, dict) and isinstance(parameter_data.get('object_id'), int) and parameter_schema is not None and parameter_schema.get('type') in ('sample', 'measurement', 'object_reference'):
                        referenced_object_id = parameter_data['object_id']
                        try:
                            if Permissions.READ not in get_user_object_permissions(referenced_object_id, user_id):
                                parameter_data = None
                                break
                            referenced_object = get_object(referenced_object_id)
                        except ObjectDoesNotExistError:
                            parameter_data = None
                            break
                        parameter_data = referenced_object.data
                        parameter_schema = referenced_object.schema
                    else:
                        parameter_data = None
                        break
                notebook_template['params'][parameter] = parameter_data
            else:
                notebook_template['params'][parameter] = None
    return notebook_templates
